// File: DGPSListener.h (to be included by DGPSListener.cpp)

// Written by Walter Piechulla in April 2000

// Sam Blackburn's prologue about Geodesy Foundation Classes:
// http://ourworld.compuserve.com/homepages/Sam_Blackburn/homepage.htm

// Definitions
// Ellipsoid
//  A flattened sphere. Take a basketball (a sphere), let some air out of it then
//  stand on it. You now have an ellipsoid. The basketball is flat on the bottom
//  and the top but still round in the middle.
// Equator
//  0 degrees Latitude
// Flattening
//  This is a ratio between polar radius and equatorial radius. Other names for
//  flattening are elliptocyte or oblateness.
// Latitude
//  The lines that run around the Earth like belts. They measure up/down angles.
// Longituide
//  The lines that slice the Earth like an orange from top to bottom. They
//  measure left/right angles.
// Meridian
//   One of the imaginary circles on the earth's surface passing through the north
//   and south poles 
// Pi
//  The most famous constant. It is a ratio.
//  It is the Circumference of a circle divided by
//  the diameter of the circle. It is 3.14 (or roughly 22 divivded by 7).
// Prime Meridian 
//  0 degrees Longitude
// Radian
//  The unit of measurement of a circle. It is a ratio.
//  It's kind of hard to explain what
//  a radian is so let me give you an example. Take the basketball and cut
//  a length of string equal to the radius (i.e. half the width) of the
//  ball. Now, lay this string on top of the ball. See how it forms an arc?
//  Now, draw a line from each end of the string to the center of the ball.
//  The angle of these two lines at the center of the ball is roughly
//  57.2958 degrees. This is true no matter what the size of the ball is.
//  Why? Because as the size of the ball increases, so does the radius.
//  So, a radian can be considered as a relationship between the radius of
//  a sphere (ball) and the circumference of the sphere. Now, there's something
//  interesting about the number of degrees in a radian (52.2958). It is
//  equal to 180 divided by Pi. Yup! Go Figure! OK. Now we're getting somewhere.
//  To find the length of the arc when all you have is the number of radians
//  in the angle at the center of the sphere, you multiply the radius times
//  the number of radians. It's that simple (on a perfect sphere anyway).

// Geodetic Datum - description of the shape of the earth

// Reference Ellipsoid - A flattened sphere. Let some air out of
// a basketball, then stand on it. The ball is now an ellipsoid.
// Your feet are standing on the North Pole and the floor is the
// South Pole.

// Ellipsoids are described by their polar and equatorial radii.
// Polar radius is also known as semi-minor axis.
// Equatorial radius is also know as semi-major axis.
// All other items of interest about the ellipsoid are derived
// from these two data.
// Flattening is ( Equatorial radius - Polar radius ) / Equatorial radius
// There's another thing called First Eccentricity Squared, this is computed
// by ( 2 * Flattening ) - ( Flattening squared ).

// Coordinates - a means of expressing a point in space
//
// Cartesian coordinates are X, Y, and Z coordinates. These are always
// expressed in distances from 0, 0, 0 (i,e, the center of the earth).
//
// Polar coordinates are theta (often seen as a letter O with a line
// through its middle running left-to-right), phi (often seen as a
// letter O with a line through its middle running top-to-bottom), and
// r (a distance).
// These are two angles and a distance. The angles are measured from
// a common origin (the center of the earth). Theta is the plane that
// runs through the equator, phi runs through the poles. R is the 
// distance along the line running along the phi angle. For simplicity
// sake, we will refer to theta as Equatorial angle, phi as the
// polar angle, and r as Polar Distance.
//
// Converting coordinates
//
// You can convert from polar to cartesian cordinates using the following
// formula:
// X = Polar distance * cosine of Polar angle * cosine of Equatorial angle
// Y = Polar distance * cosine of Polar angle * sine of Equatorial angle
// Z = Polar distance * sine of Polar angle

// Applying this to the real world
//
// Cartesian coordinates ar commonly refered to an ECEF X,Y,Z coordinates.
// This means Earth Centered, Earth Fixed. The 0,0,0 coordinate is the
// center of the earth. The X axis runs towards the Prime Meridian.
// The Y axis runs towards the equator and the Z axis runs toward
// the poles. Positive Z numbers go towards the North pole while
// negative numbers go towards the South Pole. Positive 

// Computing Distance
//
// If you have two cartesian coordinates, you can compute a line
// of sight (as the bullet flies, aiming a laser, pointing in a straight line)
// by this formula (some guy named Pythagoras figured this out):
// SQRT( ( X1 - X2 ) ^ 2 + ( Y1 - Y2 ) ^ 2 + ( Z1 - Z2 ) ^ 2 )
//
// or in pseudo code:
//
// cartesian_coordinate first_location;
// cartesian_coordinate second_location;
//
// double temp_x;
// double temp_y;
// double temp_z;
//
// temp_x = first_location.X - second_location.X;
// temp_y = first_location.Y - second_location.Y;
// temp_z = first_location.Z - second_location.Z;
//
// temp_x = temp_x * temp_x; // square them
// temp_y = temp_y * temp_y;
// temp_z = temp_z * temp_z;
//
// double temp_double;
//
// temp_double = temp_x + temp_y + temp_z;
//
// double distance;
//
// distance = sqrt( temp_double );
//
// While this will give you distance, it will not give you direction.

// End of Sam Blackburn's prologue

#include <math.h>

#define IDR_MTTTYMENU                   101
#define IDR_MTTTYACCELERATOR            102
#define IDI_APPICON                     103
#define IDD_TOOLBARSETTINGS             103
#define ID_TTYWINDOW                    103
#define IDD_ABOUT                       104
#define IDD_STATUSDIALOG                105
#define IDD_COMMEVENTSDLG               107
#define IDD_FLOWCONTROLDLG              108
#define IDI_APPICON2                    109
#define IDI_APPICON3                    110
#define IDD_TIMEOUTSDLG                 110
#define IDI_APPICON4                    111
#define IDD_GETADWORD                   111
#define IDC_PORTCOMBO                   1000
#define IDC_BAUDCOMBO                   1001
#define IDC_PARITYCOMBO                 1002
#define IDC_DATABITSCOMBO               1003
#define IDC_STOPBITSCOMBO               1004
#define IDC_STATCTS                     1004
#define IDC_STATDSR                     1005
#define IDC_STATRING                    1006
#define IDC_STATRLSD                    1007
#define IDC_LOCALECHOCHK                1008
#define IDC_DISPLAYERRORSCHK            1009
#define IDC_FONTBTN                     1010
#define IDC_ABORTBTN                    1011
#define IDC_TRANSFERPROGRESS            1013
#define IDC_MODEMSTATUSGRP              1018
#define IDC_COMMEVENTSBTN               1019
#define IDC_EVBREAKBTN                  1020
#define IDC_EVCTSBTN                    1021
#define IDC_EVDSRBTN                    1022
#define IDC_EVERRBTN                    1023
#define IDC_EVRINGBTN                   1024
#define IDC_EVRLSDBTN                   1025
#define IDC_EVRXCHARBTN                 1026
#define IDC_EVRXFLAGBTN                 1027
#define IDC_EVTXEMPTYBTN                1028
#define IDC_FLAGEDIT                    1029
#define IDC_DEFAULTSBTN                 1030
#define IDC_LFBTN                       1032
#define IDC_AUTOWRAPCHK                 1033
#define IDC_STATUSEDIT                  1034
#define IDC_FLAGCHAR                    1036
#define IDC_FLOWCONTROLBTN              1039
#define IDC_CTSOUTCHK                   1040
#define IDC_DSROUTCHK                   1041
#define IDC_DTRCONTROLCOMBO             1043
#define IDC_DSRINCHK                    1044
#define IDC_XONXOFFOUTCHK               1045
#define IDC_XONXOFFINCHK                1046
#define IDC_XONLIMITEDIT                1047
#define IDC_XOFFLIMITEDIT               1048
#define IDC_XONCHAREDIT                 1049
#define IDC_XOFFCHAREDIT                1050
#define IDC_RTSCONTROLCOMBO             1051
#define IDC_RTSCTSBTN                   1052
#define IDC_XOFFXONBTN                  1053
#define IDC_XONCHARDISP                 1054
#define IDC_XOFFCHARDISP                1055
#define IDC_PICTURE                     1056
#define IDC_DTRDSRBTN                   1056
#define IDC_TIMEOUTSBTN                 1057
#define IDC_READINTERVALEDIT            1058
#define IDC_READMULTIPLIEREDIT          1059
#define IDC_READCONSTANTEDIT            1060
#define IDC_WRITEMULTIPLIEREDIT         1061
#define IDC_WRITECONSTANTEDIT           1062
#define IDC_NONEBTN                     1064
#define IDC_OSVERSIONINFO               1065
#define IDC_TXAFTERXOFFSENTCHK          1066
#define IDC_CTSHOLDCHK                  1067
#define IDC_DSRHOLDCHK                  1068
#define IDC_RLSDHOLDCHK                 1069
#define IDC_XOFFHOLDCHK                 1070
#define IDC_XOFFSENTCHK                 1071
#define IDC_EOFSENTCHK                  1072
#define IDC_TXIMCHK                     1073
#define IDC_TXCHAREDIT                  1074
#define IDC_RXCHAREDIT                  1075
#define IDC_NOREADINGCHK                1076
#define IDC_NOWRITINGCHK                1077
#define IDC_NOSTATUSCHK                 1078
#define IDC_NOEVENTSCHK                 1079
#define IDC_DWORDSTATIC                 1081
#define IDC_DWORDEDIT                   1082
#define IDC_DISPLAYTIMEOUTS             1083
#define ID_FILE_EXIT                    40001
#define ID_HELP_ABOUTMTTTY              40002
#define ID_FILE_CONNECT                 40004
#define ID_FILE_DISCONNECT              40005
#define ID_TTY_CLEAR                    40008
#define ID_TRANSFER_SENDFILETEXT        40010
#define ID_TRANSFER_RECEIVEFILETEXT     40011
#define ID_TRANSFER_SENDREPEATEDLY      40012
#define ID_TRANSFER_ABORTSENDING        40016
#define ID_TRANSFER_ABORTREPEATEDSENDING 40018
#define ID_BLACKBOX_DUMPRAW             40020 
#define ID_BLACKBOX_DUMPCOOKED          40021
#define ID_BLACKBOX_STOPSCREENDUMP      40022
#define ID_BLACKBOX_CAPTURERAW          40023
#define ID_BLACKBOX_CAPTURECOOKED       40024
#define ID_BLACKBOX_GPALM_ENABLED       40050 // consecutive numbering because of checkmarks 
#define ID_BLACKBOX_GPGGA_ENABLED       40051
#define ID_BLACKBOX_GPGSA_ENABLED       40052 
#define ID_BLACKBOX_GPGSV_ENABLED       40053
#define ID_BLACKBOX_GPRMC_ENABLED       40054
#define ID_BLACKBOX_GPVTG_ENABLED       40055
#define ID_BLACKBOX_LCGLL_ENABLED       40056
#define ID_BLACKBOX_LCVTG_ENABLED       40057
#define ID_BLACKBOX_PGRME_ENABLED       40058
#define ID_BLACKBOX_PGRMF_ENABLED       40059
#define ID_BLACKBOX_PGRMT_ENABLED       40060
#define ID_BLACKBOX_PGRMV_ENABLED       40061 
#define ID_BLACKBOX_ENABLE_ALL          40075
#define ID_BLACKBOX_DISABLE_ALL         40076  
#define ID_BLACKBOX_SETBEGINMARKER      40077
#define ID_BLACKBOX_SETENDMARKER        40078 
#define IDC_STATIC                      65535

// constant definitions

// hard coded maximum number of ports
#define MAXPORTS        4

// terminal size
#define MAXROWS         50
#define MAXCOLS         80

// cursor states
#define CS_HIDE         0x00
#define CS_SHOW         0x01

// ascii definitions
#define ASCII_BEL       0x07
#define ASCII_BS        0x08
#define ASCII_LF        0x0A
#define ASCII_CR        0x0D
#define ASCII_XON       0x11
#define ASCII_XOFF      0x13

// data structures
struct TTYInfoStruct
{
  HANDLE  hCommPort, hReaderStatus, hWriter ;
  DWORD   dwEventFlags;
  CHAR    Screen[MAXCOLS * MAXROWS];
  CHAR    chFlag, chXON, chXOFF;
  WORD    wXONLimit, wXOFFLimit;
  DWORD   fRtsControl;
  DWORD   fDtrControl;
  BOOL    fConnected, fTransferring, fRepeating,
          fLocalEcho, fNewLine,
          fDisplayErrors, fAutowrap,
          fCTSOutFlow, fDSROutFlow, fDSRInFlow, 
          fXonXoffOutFlow, fXonXoffInFlow,
          fTXafterXoffSent,
          fNoReading, fNoWriting, fNoEvents, fNoStatus,
          fDisplayTimeouts;
  BYTE    bPort, bByteSize, bParity, bStopBits ;
  DWORD   dwBaudRate ;
  WORD    wCursorState ;
  HFONT   hTTYFont ;
  LOGFONT lfTTYFont ;
  DWORD   rgbFGColor ;
  COMMTIMEOUTS timeoutsorig;
  COMMTIMEOUTS timeoutsnew;
  int     xSize, ySize, xScroll, yScroll, xOffset, yOffset,
          nColumn, nRow, xChar, yChar , nCharPos;

} TTYInfo;


// macros ( for easier readability )
#define COMDEV( x )         (x.hCommPort)
#define CURSORSTATE( x )    (x.wCursorState)
#define PORT( x )           (x.bPort)
#define SCREEN( x )         (x.Screen)
#define CONNECTED( x )      (x.fConnected)
#define TRANSFERRING( x )   (x.fTransferring)
#define REPEATING( x )      (x.fRepeating)
#define LOCALECHO( x )      (x.fLocalEcho)
#define NEWLINE( x )        (x.fNewLine)
#define AUTOWRAP( x )       (x.fAutowrap)
#define BYTESIZE( x )       (x.bByteSize)
#define PARITY( x )         (x.bParity)
#define STOPBITS( x )       (x.bStopBits)
#define BAUDRATE( x )       (x.dwBaudRate)
#define HTTYFONT( x )       (x.hTTYFont)
#define LFTTYFONT( x )      (x.lfTTYFont)
#define FGCOLOR( x )        (x.rgbFGColor)
#define XSIZE( x )          (x.xSize)
#define YSIZE( x )          (x.ySize)
#define XSCROLL( x )        (x.xScroll)
#define YSCROLL( x )        (x.yScroll)
#define XOFFSET( x )        (x.xOffset)
#define YOFFSET( x )        (x.yOffset)
#define COLUMN( x )         (x.nColumn)
#define ROW( x )            (x.nRow)
#define XCHAR( x )          (x.xChar)
#define YCHAR( x )          (x.yChar)
#define DISPLAYERRORS( x )  (x.fDisplayErrors)
#define TIMEOUTSORIG( x )   (x.timeoutsorig)
#define TIMEOUTSNEW( x )    (x.timeoutsnew)
#define WRITERTHREAD( x )   (x.hWriter)
#define READSTATTHREAD( x ) (x.hReaderStatus)
#define EVENTFLAGS( x )     (x.dwEventFlags)
#define FLAGCHAR( x )       (x.chFlag)
#define SCREENCHAR( x, col, row )   (x.Screen[row * MAXCOLS + col])

#define DTRCONTROL( x )     (x.fDtrControl)
#define RTSCONTROL( x )     (x.fRtsControl)
#define XONCHAR( x )        (x.chXON)
#define XOFFCHAR( x )       (x.chXOFF)
#define XONLIMIT( x )       (x.wXONLimit)
#define XOFFLIMIT( x )      (x.wXOFFLimit)
#define CTSOUTFLOW( x )     (x.fCTSOutFlow)
#define DSROUTFLOW( x )     (x.fDSROutFlow)
#define DSRINFLOW( x )      (x.fDSRInFlow)
#define XONXOFFOUTFLOW( x ) (x.fXonXoffOutFlow)
#define XONXOFFINFLOW( x )  (x.fXonXoffInFlow)
#define TXAFTERXOFFSENT(x)  (x.fTXafterXoffSent)

#define NOREADING( x )      (x.fNoReading)
#define NOWRITING( x )      (x.fNoWriting)
#define NOEVENTS( x )       (x.fNoEvents)
#define NOSTATUS( x )       (x.fNoStatus)
#define SHOWTIMEOUTS( x )   (x.fDisplayTimeouts)
// end of #include "ttyinfo.h"

// GLOBAL DEFINES
#define TTY_BUFFER_SIZE         MAXROWS * MAXCOLS
#define MAX_STATUS_BUFFER       20000
#define MAX_WRITE_BUFFER        1024
#define MAX_READ_BUFFER         2048
#define READ_TIMEOUT            500
#define STATUS_CHECK_TIMEOUT    500
#define WRITE_CHECK_TIMEOUT     500
#define PURGE_FLAGS             PURGE_TXABORT | PURGE_TXCLEAR | PURGE_RXABORT | PURGE_RXCLEAR 
#define EVENTFLAGS_DEFAULT      EV_BREAK | EV_CTS | EV_DSR | EV_ERR | EV_RING | EV_RLSD
#define FLAGCHAR_DEFAULT        '\n'

// Write request types
#define WRITE_CHAR          0x01
#define WRITE_FILE          0x02
#define WRITE_FILESTART     0x03
#define WRITE_FILEEND       0x04
#define WRITE_ABORT         0x05
#define WRITE_BLOCK         0x06

// Read states
#define RECEIVE_TTY                    0x01
#define RECEIVE_CAPTURED               0x02
#define RECEIVE_TTYBLACKBOXCOOKED      0x03
#define RECEIVE_CAPTUREDBLACKBOXCOOKED 0x04

// window coords
#define MAXXWINDOW          750
#define MAXYWINDOW          530
#define STARTXWINDOW        80
#define STARTYWINDOW        70

#define SETTINGSFACTOR      5
#define STATUSFACTOR        5

// window timer ids
#define TIMERID             1

#define MAXLEN_TEMPSTR      20
#define MAX_STATUS_LENGTH   100

#define AMOUNT_TO_READ          512
#define NUM_READSTAT_HANDLES    4

// GLOBAL VARIABLES
OSVERSIONINFO gOSV;
HINSTANCE     ghInst;
HACCEL        ghAccel;
HWND          ghwndMain;
HWND          ghWndToolbarDlg;
HWND          ghWndStatusDlg;
HWND          ghWndTTY;
HWND          ghWndHidden;

//  Window placement variables
WORD gwBaseY;
LONG gcyMinimumWindowHeight;

//  Flags controlling thread actions
HANDLE ghThreadExitEvent;
BOOL gfAbortTransfer;

//  File transfer variables
DWORD  gdwFileTransferLeft;
DWORD  gdwReceiveState;
HANDLE ghFileCapture;

//  Status updating
CRITICAL_SECTION gStatusCritical;
HANDLE ghStatusMessageEvent;
HANDLE ghStatusMessageHeap;
HFONT ghFontStatus;
int   gnStatusIndex;

typedef struct STATUS_MESSAGE;

struct STATUS_MESSAGE * glpStatusMessageHead;
struct STATUS_MESSAGE * glpStatusMessageTail;

typedef struct STATUS_MESSAGE
{
  struct STATUS_MESSAGE * lpNext;     // pointer to next node
  char chMessageStart;                // variable length string start here
} STATUS_MESSAGE;

//  Port name
char gszPort[10];

//  Writer heap variables
CRITICAL_SECTION gcsWriterHeap;
CRITICAL_SECTION gcsDataHeap;
HANDLE ghWriterHeap;
HANDLE ghWriterEvent;
HANDLE ghTransferCompleteEvent;

//  Write request data structure; look in Writer.c for more info
typedef struct WRITEREQUEST;

struct WRITEREQUEST *gpWriterHead;
struct WRITEREQUEST *gpWriterTail;

typedef struct WRITEREQUEST
{
  DWORD      dwWriteType;        // char, file start, file abort, file packet
  DWORD      dwSize;             // size of buffer
  char       ch;                 // ch to send
  char *     lpBuf;              // address of buffer to send
  HANDLE     hHeap;              // heap containing buffer
  HWND       hWndProgress;       // status bar window handle
  struct WRITEREQUEST *pNext;    // next node in the list
  struct WRITEREQUEST *pPrev;    // prev node in the list
} WRITEREQUEST, *PWRITEREQUEST;

// Variables from the original file transfer.c
HANDLE hFile;
HANDLE hTransferAbortEvent;
HANDLE hTransferThread;
UINT   uTimerId;
MMRESULT mmTimer = (MMRESULT)NULL;
char * lpBuf;

// BlackBoxPlus specific globals
char CmdBuffer[256]; // buffer is even oversized, since no NMEA 0183 command has more than 80 chars
BOOL BeginMarkerWaitingToBeSet;
BOOL EndMarkerWaitingToBeSet;

// For Walter's minimal NMEA-Parser

#if ! defined(CARRIAGE_RETURN)
#define CARRIAGE_RETURN 0x0D
#endif

#if ! defined(LINE_FEED)
#define LINE_FEED       0x0A
#endif

typedef enum
{
  UNIMPLEMENTED = 0,
  GPGGA,
  GPGSA,
  GPGSV,
  GPRMC,
  PGRME,
  PGRMT
} KNOWNSENTENCES;

enum 
{
  CHECKSUM_PARSE_ERROR = 11,
  UNKNOWNSENTENCE_PARSE_ERROR
};

// For selection / deselection of Garmin sentences
typedef enum
{
  S_GPALM = 0,
  S_GPGGA,
  S_GPGSA,
  S_GPGSV,
  S_GPRMC,
  S_GPVTG,
  S_LCGLL,
  S_LCVTG,
  S_PGRME,
  S_PGRMF,
  S_PGRMT,
  S_PGRMV,
  S_IGNORED
} BBSENTENCE;

typedef enum
{
  SELECTALL = 0,
  DESELECTALL,
  SELECT,
  DESELECT
} BBACTION;

// Prototypes of functions
BOOL VersionCheck(void);
BOOL InitializeApp(HINSTANCE,int);
void BlackBoxDGPSSSetSentencesMenuChecks(HWND);
int WINAPI MTTTYWndProc(HWND,UINT,WPARAM,LPARAM);
void CmdDispatch(int,HWND,LPARAM);
void OpenTTYChildWindow(HWND);
BOOL ScrollTTYVert(HWND,WORD,WORD);
BOOL ScrollTTYHorz(HWND,WORD,WORD);
BOOL PaintTTY(HWND);
BOOL NEAR MoveTTYCursor(HWND);
BOOL NEAR SetTTYFocus(HWND);
BOOL NEAR KillTTYFocus(HWND);
BOOL NEAR SizeTTY(HWND,WORD,WORD);
int WINAPI TTYChildProc(HWND,UINT,WPARAM,LPARAM);
void GlobalInitialize(void);
void GlobalCleanup(void);
BOOL ClearTTYContents(void);
void InitNewFont(LOGFONT,COLORREF);
BOOL InitTTYInfo(void);
void DestroyTTYInfo(void);
void StartThreads(void);
HANDLE SetupCommPort(void);
DWORD WaitForThreads(DWORD);
BOOL BreakDownCommPort(void);
BOOL DisconnectOK(void);
void TransferRepeatCreate(LPCTSTR,DWORD);
void TransferRepeatDestroy(void);
void CALLBACK TransferRepeatDo(UINT,UINT,DWORD,DWORD,DWORD);
void TransferFileTextStart(LPCTSTR);
void TransferFileTextEnd(void);
void ReceiveFileText(LPCTSTR,BOOL);
HANDLE OpenTheFile(LPCTSTR);
HANDLE CreateTheFile(LPCTSTR);
BOOL GetTransferSizes(HANDLE,DWORD*,DWORD*,DWORD*);
void ShowTransferStatistics(DWORD,DWORD,DWORD);
UINT CheckForMessages(void);
void CaptureFile(HANDLE,HWND,BOOL);
DWORD WINAPI TransferThreadProc(LPVOID);
void OpenSettingsToolbar(HWND);
void ChangeConnection(HWND,BOOL);
void UpdateTTYInfo(void);
BOOL UpdateConnection(void);
void FillComboBox(HWND,char **,DWORD *,WORD,DWORD);
void SetComboBox(HWND,WORD,DWORD);
BOOL SettingsDlgInit(HWND);
DWORD GetdwTTYItem(HWND,int,char **,DWORD *,int);
BYTE GetbTTYItem(HWND,int, char **,DWORD *,int);
BOOL CALLBACK ToolbarProc(HWND,UINT,WPARAM,LPARAM);
void InitHexControl(HWND,WORD,WORD,char);
char GetHexControl(HWND,WORD,WORD);
void InitCommEventsDlg(HWND,DWORD);
void SaveCommEventsDlg(HWND);
BOOL CALLBACK CommEventsProc(HWND,UINT,WPARAM,LPARAM);
void SaveFlowControlDlg(HWND);
void InitFlowControlDlg(HWND);
void FlowDefault(HWND,WORD);
BOOL CALLBACK FlowControlProc(HWND,UINT,WPARAM,LPARAM);
void InitTimeoutsDlg(HWND,COMMTIMEOUTS);
void SaveTimeoutsDlg(HWND);
BOOL CALLBACK TimeoutsProc(HWND,UINT,WPARAM,LPARAM);
BOOL CALLBACK GetADWORDProc(HWND,UINT,WPARAM,LPARAM);
DWORD GetAFrequency(void);
void OpenStatusToolbar(HWND);
HFONT CreateStatusEditFont(void);
BOOL CALLBACK StatusDlgProc(HWND,UINT,WPARAM,LPARAM);
void InitStatusMessage(void);
void StatusMessage(void);
void UpdateStatus(char *);
void ReportModemStatus(DWORD);
void CheckModemStatus(BOOL);
void ReportComStat(COMSTAT);
void CheckComStat(BOOL);
void ReportStatusEvent(DWORD);
BOOL CmdAbout(HWND);
UINT InitAboutDlg(HWND);
BOOL CALLBACK AboutDlgProc(HWND,UINT,WPARAM,LPARAM);
DWORD ErrorExtender(DWORD,char **);
void ErrorReporter(char *);
void OutputABufferToWindow(HWND,char *,DWORD);
void OutputABufferToWindow(HWND,char *,DWORD);
void OutputABufferToFile(HANDLE,char *,DWORD);
void OutputABufferToFile(HANDLE,char *,DWORD);
void OutputABuffer(HWND,char *,DWORD);
DWORD WINAPI WriterProc(LPVOID);
void HandleWriteRequests();
void WriterComplete(void);
void WriterAbort(PWRITEREQUEST);
void WriterFile(PWRITEREQUEST);
void WriterBlock(PWRITEREQUEST);
void WriterFileStart(DWORD);
void WriterChar(PWRITEREQUEST);
void WriterGeneric(char *,DWORD);
BOOL WriterAddNewNode(DWORD,DWORD,char,char *,HANDLE,HWND);
BOOL WriterAddNewNodeTimeout(DWORD,DWORD,char,char *,HANDLE,HWND,DWORD);
BOOL WriterAddFirstNodeTimeout(DWORD,DWORD,char,char *,HANDLE,HWND,DWORD);
BOOL WriterAddExistingNode(PWRITEREQUEST,DWORD,DWORD,char,char *,HANDLE,HWND);
void AddToLinkedList(PWRITEREQUEST);
void AddToFrontOfLinkedList(PWRITEREQUEST);
PWRITEREQUEST RemoveFromLinkedList(PWRITEREQUEST);
void ErrorInComm(char *);
DWORD WINAPI ReaderAndStatusProc(LPVOID);
void ErrorHandler(char *);
// BlackBoxPlus specific functions
BOOL BlackBoxDGPSSelectSentences(BBSENTENCE,BBACTION);
void BlackBoxDGPSInterpretedWindowOutput(HWND,char *,DWORD);
void BlackBoxDGPSInterpretedFileOutput(HANDLE,char *,DWORD);
BOOL BlackBoxEndOfCommand(char *,DWORD,char *,char *);
void BlackBoxInterpretCmd(char *);
// Minimal NMEA Parser functions
KNOWNSENTENCES NMEAParserGetSentenceType(const char *);
DWORD NMEAParserComputeChecksum(char *);
DWORD NMEAParserGetContainedChecksum(char *);
DWORD NMEAParserHexValue(const char *);
void NMEAParserAnalyzeGPRMC(char *);
void NMEAParserAnalyzeGPGGA(char *);
CString NMEAParserField(int,char *);
static inline double NMEAParserGetDegreesFromNMEALat(const char,const char [10]);
static inline double NMEAParserGetMinutesFromNMEALat(const char [10]);
static inline double NMEAParserGetDegreesFromNMEALon(const char,const char [11]);
static inline double NMEAParserGetMinutesFromNMEALon(const char [11]);

// Inline functions definitions

// Remember: North is positive, south is negative
inline double NMEAParserGetDegreesFromNMEALat(const char nOrS,const char lat[10])
{
  char buf[3];
  double result;

  buf[0] = lat[0];
  buf[1] = lat[1];
  buf[2] = '\0';

  result = atof(buf); 

  if (nOrS == 'S') // NMEA uses capital letters
  {
    result = -result;
  }
  
  return result; 
}

inline double NMEAParserGetMinutesFromNMEALat(const char lat[10])
{
  char buf[8];

  buf[0] = lat[2];
  buf[1] = lat[3];  
  buf[2] = lat[4];
  buf[3] = lat[5];
  buf[4] = lat[6];
  buf[5] = lat[7];
  buf[6] = lat[8];
  buf[7] = '\0';

  return(atof(buf)); 
}

// Remember: West is negative, East is positive
inline double NMEAParserGetDegreesFromNMEALon(const char eOrW,const char lon[11])
{
  char buf[4];
  double result;

  buf[0] = lon[0];
  buf[1] = lon[1];
  buf[2] = lon[2];
  buf[3] = '\0';

  result = atof(buf);
  
  if (eOrW == 'W')  // NMEA uses capital letters
  {
    result = -result;
  }
  
  return result; 
}

static inline double NMEAParserGetMinutesFromNMEALon(const char lon[11])
{
  char buf[8];

  buf[0] = lon[3];
  buf[1] = lon[4];  
  buf[2] = lon[5];
  buf[3] = lon[6];
  buf[4] = lon[7];
  buf[5] = lon[8];
  buf[6] = lon[9];
  buf[7] = '\0';

  return(atof(buf)); 
}

// WGS84 to Gauss-Krueger converter (recoding of Pascal sources from Ottmar Labonde)

class WGS84toDHDN
{
protected:
  void HelmertTransformation(double,double,double,double&,double&,double&);
  void BesselBLnachGaussKrueger(double,double,double&,double&);
  void BLRauenberg (double,double,double,double&,double&,double&);
  double neuF(double,double,double,double);
  double round(double);

private:
  double awgs;  // WGS84 Semi-Major Axis = Equatorial Radius in meters
  double bwgs;  // WGS84 Semi-Minor Axis = Polar Radius in meters 
  double abes;  // Bessel Semi-Major Axis = Equatorial Radius in meters
  double bbes;  // Bessel Semi-Minor Axis = Polar Radius in meters
  double cbes;  // Bessel ???
  double dx;    // Translation Parameter 1
  double dy;    // Translation Parameter 2 
  double dz;    // Translation Parameter 3
  double rotx;  // Rotation Parameter 1
  double roty;  // Rotation Parameter 2
  double rotz;  // Rotation Parameter 3
  double sc;    // Scaling Factor 
  double Pi;    // The famous number
  double l1;    // Helper          
  double b1;    // Helper  
  double l2;    // Helper
  double b2;    // Helper
  double h1;    // Helper
  double h2;    // Helper  
  double R;     // Helper
  double H;     // Helper
  double a;     // Helper    
  double b;     // Helper  
  double eq;    // Helper
  double N;     // Helper    
  double Xq;    // Helper
  double Yq;    // Helper
  double Zq;    // Helper  
  double X;     // Helper
  double Y;     // Helper
  double Z;     // Helper

public:
  WGS84toDHDN();
  ~WGS84toDHDN();
  void Convert(double,double,double,double&,double&);
};

// Geodesy Foundation Classes (Sam Blackburn)

class CMath
{
  // This class encapsulates all of the math functions. It is here to allow you
  // to replace the C Runtime functions with your home-grown (and maybe better
  // implementation) routines

public:

  static inline double AbsoluteValue( const double& value );
  static inline double ArcCosine( const double& value );
  static inline double ArcSine( const double& value );
  static inline double ArcTangent( const double& value );
  static inline double ArcTangentOfYOverX( const double& y, const double& x );
  static inline double Ceiling( const double& value );
  static inline double ConvertDegreesToRadians( const double& degrees );
  static inline double ConvertRadiansToDegrees( const double& radians );
  static inline void   ConvertDecimalDegreesToDegreesMinutesSeconds( double decimal_degrees, double& degrees, double& minutes, double& seconds );
  // West is negative, East is positive, North is positive, south is negative
  static inline double ConvertDegreesMinutesSecondsCoordinateToDecimalDegrees( double degrees, double minutes, double seconds );
  static inline double Cosine( const double& value );
  static inline double HyperbolicCosine( const double& value );
  static inline double Pi( void );
  static inline double Sine( const double& value );
  static inline double SquareRoot( const double& value );
};

// This is Blackburn's inline functions implementation (original "CMath.inl") 

inline double CMath::AbsoluteValue(const double& value)
{
  return(::fabs(value));
}

inline double CMath::ArcCosine(const double& value)
{
  return(::acos(value));
}

inline double CMath::ArcSine(const double& value)
{
  return(::asin(value));
}

inline double CMath::ArcTangent(const double& value)
{
  return(::atan(value));
}

inline double CMath::ArcTangentOfYOverX( const double& y, const double& x)
{
  return(::atan2(y,x));
}

inline double CMath::Ceiling(const double& value)
{
  return(::ceil(value));
}

inline void CMath::ConvertDecimalDegreesToDegreesMinutesSeconds(double decimal_degrees,
                                                                double& degrees,
                                                                double& minutes,
                                                                double& seconds)
{
  double fractional_part = 0.0;

  double integer_part = 0;

  fractional_part = ::modf(decimal_degrees,&integer_part);

  degrees = integer_part;

  if (decimal_degrees < 0.0)
  {
    fractional_part *= (-1.0);
  }

  minutes = fractional_part * 60.0;

  fractional_part = ::modf(minutes,&integer_part);

  minutes = integer_part;

  seconds = fractional_part * 60.0;
}

inline double CMath::ConvertDegreesMinutesSecondsCoordinateToDecimalDegrees(double degrees, 
                                                                            double minutes, 
                                                                            double seconds)
{
  double decimal_degrees = 0.0;

  decimal_degrees = degrees;

  if (decimal_degrees < 0.0)
  {
    decimal_degrees *= (-1.0);
  }

  decimal_degrees += (double) (minutes / 60.0);
  decimal_degrees += (double) (seconds / 3600.0);

  if ( degrees < 0.0 )
  {
    decimal_degrees *= (-1.0);
  }

  return(decimal_degrees);
}

inline double CMath::ConvertDegreesToRadians(const double& degrees)
{
  double radians = 0.0;
  double pi_divided_by_180 = CMath::Pi() / 180.0;
   
  radians = degrees * pi_divided_by_180;

  return( radians );
}

inline double CMath::ConvertRadiansToDegrees(const double& radians)
{
  double degrees = 0.0;

  double conversion_factor = 180.0 / CMath::Pi();

  degrees = radians * conversion_factor;

  return( degrees );
}

inline double CMath::Cosine(const double& value)
{
  return(::cos(value));
}

inline double CMath::HyperbolicCosine(const double& value)
{
  return(::cosh(value));
}

inline double CMath::Pi(void)
{
  return( 3.1415926535897932384626433832795028841971693993751058209749445923078164 );
}

inline double CMath::Sine(const double& value)
{
  return(::sin(value));
}

inline double CMath::SquareRoot(const double& value)
{
  return(::sqrt(value));
}

// End of the inline functions

class CEarthCoordinate
{
  // This is a Cartesian coordinate (Earth Centered, Earth Fixed)

protected:

  double m_X_CoordinateInMeters; // Positive points to intersection of the Prime Meridian and the equator
  double m_Y_CoordinateInMeters; // Positive points to the intersection of 90 degrees east of Prime Meridian and the equator
  double m_Z_CoordinateInMeters; // Positive points to the North Pole, Negative towards the South Pole

public:

  CEarthCoordinate();
  CEarthCoordinate(const CEarthCoordinate& source);
  ~CEarthCoordinate();

  void   Copy(const CEarthCoordinate& source);
  void   Get(double& x_coordinate,double& y_coordinate,double& z_coordinate) const;
  double GetXCoordinateInMeters(void) const;
  double GetYCoordinateInMeters(void) const;
  double GetZCoordinateInMeters(void) const;
  void   Set(double x_coordinate,double y_coordinate,double z_coordinate);
  void   SetXCoordinateInMeters(double x_coordinate);
  void   SetYCoordinateInMeters(double y_coordinate);
  void   SetZCoordinateInMeters(double z_coordinate);

  CEarthCoordinate& operator = (const CEarthCoordinate& source);
};

class CPolarCoordinate
{
protected:

  double m_UpDownAngleInDegrees; // Polar Angle, Phi
  double m_LeftRightAngleInDegrees; // Equatorial Angle, Theta
  double m_DistanceFromSurfaceInMeters;

public:

  CPolarCoordinate();
  CPolarCoordinate(const CPolarCoordinate& source);
  ~CPolarCoordinate();

  void   Copy(const CPolarCoordinate& source);
  void   Get(double& up_down_angle,double& left_right_angle,double& distance_from_surface) const;
  double GetUpDownAngleInDegrees(void) const;
  double GetLeftRightAngleInDegrees(void) const;
  double GetDistanceFromSurfaceInMeters(void) const;
  void   Set(double up_down_angle,double left_right_angle,double distance_from_surface);
  void   SetUpDownAngleInDegrees(double up_down_angle);
  void   SetLeftRightAngleInDegrees(double left_right_angle);
  void   SetDistanceFromSurfaceInMeters(double distance_from_surface);

  CPolarCoordinate& operator = (const CPolarCoordinate& source);
};

class CEarth
{
protected:

  // These are the magic numbers. They are the "real" data. They are facts,
  // measurements. Everything else about an ellipse is derived (or computed from)
  // these two data items.

  double m_PolarRadiusInMeters;
  double m_EquatorialRadiusInMeters;

  // Here be the things that can be derived from our hard data.
  // We compute these values using the two pieces of data that we
  // know about the ellipse.

  double m_Flattening;
  double m_EccentricitySquared;

  // Here's stuff specific to the C++ class

  int m_EllipsoidID;

  void m_ComputeFlattening( void );
  void m_ComputeEccentricitySquared( void );
  void m_Initialize( void );
      
public:

  enum _EllipsoidType
  {
    Unknown,
    Perfect_Sphere,
    Airy,
    Austrailian_National,
    Bessell_1841,
    Bessel_1841_Nambia,
    Clarke_1866,
    Clarke_1880,
    Everest,
    Fischer_1960_Mercury,
    Fischer_1968,
    GRS_1967,
    GRS_1980,
    Helmert_1906,
    Hough,
    International,
    Krassovsky,
    Modified_Airy,
    Modified_Everest,
    Modified_Fischer_1960,
    South_American_1969,
    Topex_Poseidon_Pathfinder_ITRF,
    WGS_60,
    WGS_66,
    WGS_72,
    WGS_84,
    Custom
  }
  ELLIPSOIDS;

  enum _DatumTypes
  {
    NAD_27 = Clarke_1866,
    Tokyo  = Bessell_1841,
  }
  DATUMS;

  CEarth(int ellipsoid = WGS_84); // Default constructor sets WGS 84 model
  ~CEarth();

  virtual void   AddLineOfSightDistanceAndDirectionToCoordinate(const CPolarCoordinate& point_1,double distance,double direction,CPolarCoordinate& point_2,double height_above_surface_of_point_2 = 0.0);
  virtual void   AddSurfaceDistanceAndDirectionToCoordinate(const CEarthCoordinate& point_1,double distance,double direction,CEarthCoordinate& point_2);
  virtual void   AddSurfaceDistanceAndDirectionToCoordinate(const CEarthCoordinate& point_1,double distance,double direction,CPolarCoordinate& point_2);
  virtual void   AddSurfaceDistanceAndDirectionToCoordinate(const CPolarCoordinate& point_1,double distance,double direction,CEarthCoordinate& point_2);
  virtual void   AddSurfaceDistanceAndDirectionToCoordinate(const CPolarCoordinate& point_1,double distance,double direction,CPolarCoordinate& point_2);
  virtual void   Convert(const CEarthCoordinate& cartesian_coordinate,CPolarCoordinate& polar_coordinate) const;
  virtual void   Convert(const CPolarCoordinate& polar_coordinate,CEarthCoordinate& cartesian_coordinate) const;
  virtual double GetDistanceToHorizon(const CEarthCoordinate& point_1) const;
  virtual double GetDistanceToHorizon(const CPolarCoordinate& point_1) const;
  virtual double GetEquatorialRadiusInMeters(void) const;
  virtual double GetPolarRadiusInMeters(void) const;
  virtual double GetLineOfSightDistanceFromCourse(const CEarthCoordinate& current_location,const CEarthCoordinate& point_a,const CEarthCoordinate& point_b) const;
  virtual double GetLineOfSightDistance(const CEarthCoordinate& point_1,const CEarthCoordinate& point_2) const;
  virtual double GetLineOfSightDistance(const CPolarCoordinate& point_1,const CEarthCoordinate& point_2) const;
  virtual double GetLineOfSightDistance(const CEarthCoordinate& point_1,const CPolarCoordinate& point_2) const;
  virtual double GetLineOfSightDistance(const CPolarCoordinate& point_1,const CPolarCoordinate& point_2) const;
  virtual double GetSurfaceDistance(const CEarthCoordinate& point_1,const CEarthCoordinate& point_2,double * heading_from_point_1_to_point_2 = 0,double * heading_from_point_2_to_point_1 = 0) const;
  virtual double GetSurfaceDistance(const CEarthCoordinate& point_1,const CPolarCoordinate& point_2,double * heading_from_point_1_to_point_2 = 0,double * heading_from_point_2_to_point_1 = 0) const;
  virtual double GetSurfaceDistance(const CPolarCoordinate& point_1,const CEarthCoordinate& point_2,double * heading_from_point_1_to_point_2 = 0,double * heading_from_point_2_to_point_1 = 0) const;
  virtual double GetSurfaceDistance(const CPolarCoordinate& point_1,const CPolarCoordinate& point_2,double * heading_from_point_1_to_point_2 = 0,double * heading_from_point_2_to_point_1 = 0) const;
  virtual void   SetEllipsoid(int ellipsoid);
  virtual void   SetEllipsoidByRadii(double equatorial_radius,double polar_radius);
  virtual void   SetEllipsoidByEquatorialRadiusAndFlattening(double equatorial_radius,double flattening);
};

